/*
 * Created on Jun 28, 2007 at 8:01:42 AM.
 * 
 * Copyright (c) 2009 Robert Virkus / Enough Software
 *
 * This file is part of J2ME Polish.
 *
 * J2ME Polish is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * J2ME Polish is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with J2ME Polish; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Commercial licenses are also available, please
 * refer to the accompanying LICENSE.txt or visit
 * http://www.j2mepolish.org for details.
 */
package zip;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Vector;
import javax.microedition.io.Connector;
import javax.microedition.io.file.FileConnection;

/**
 * <p>Generates GZIP or DEFLATE encoded input streams from an InputStream.</p>
 *
 * <p>Copyright Enough Software 2007 - 2009</p>
 * <pre>
 * history
 *        Jun 28, 2007 - Simon creation
 * </pre>
 * @author Simon Schmitt, simon.schmitt@enough.de
 */
public final class ZipOutputStream {

    /**
     * This constant triggers the normal deflate compression as described in rfc 1951.
     */
    //   public static final int TYPE_DEFLATE = 0;
    /**
     * This constant triggers the gzip compression that is the same as deflate with 
     * some extra header information (see rfc 1952).
     */
    private static final int TYPE_GZIP = 1;
    private byte[] outputWindow;
    private byte[] plainDataWindow;
    private int outProcessed;
    private int plainPointer;
    private final static int HASHMAP_COUNT = 4;
    IntShortHashMap[] HM = new IntShortHashMap[HASHMAP_COUNT + 1];
    private byte[] inputBuffer;	// the input data is read in chunks from the inputStream
    private int inEnd;
    private int inStart;
    private int[] smallCodeBuffer; // we have to buffer the output, because it comes in different bitlength
    int[] huffmanCode;			// all necessary informations to store the trees
    byte[] huffmanCodeLength;
    int[] distHuffCode;
    byte[] distHuffCodeLength;
    private int[] litCount;		// the frequency of all literal symbols
    private int[] distCount;	// the frequency of all distance symbols
    // private int isize;			// # of all processed bytes from inputStream
    private int status;
    private final static int STREAM_INIT = 0;
    private final static int STREAMING = 4;
    private boolean lastBlock;	// needed to set the lastBlock bit in the header
    private boolean lz77active; // activates the lz77 compression
    private int BTYPE;	// 	=1 for fixed Tree this happens in case of huffmanwindowsize=0
    //	=2 in case of dynamix Trees
    private final static boolean debug = false;
    private final static boolean GZIP = false;
    private final static int[] crctable = new int[256];
    private boolean zipFile;
    private FileConnection fc;

    static {
        int[] ctab = crctable;
        for (int n = 0; n < 256; n++) {
            int c = n;
            for (int k = 8; --k >= 0;) {
                if ((c & 1) != 0) {
                    c = 0xedb88320 ^ (c >>> 1);
                } else {
                    c >>>= 1;
                }
            }
            ctab[n] = c;
        }
    }

     static int updateCRC(int crc, byte buf[], int off, int len) {
        int c = ~crc;
        int d[] = crctable;
        while (--len >= 0) {
            c = d[(c ^ buf[off++]) & 0xff] ^ (c >>> 8);
        }
        return ~c;
    }

    public ZipOutputStream(String url) throws IOException {
        FileConnection c = null;
        boolean ok = false;
        try {
            c = (FileConnection) Connector.open(url, Connector.WRITE);
            if (c != null) {
                try {
                    c.truncate(0);
                    ok = true;
                } catch (Throwable e) {
                    try {
                        c.create();
                        ok = true;
                    } catch (Throwable ee) {
                    }
                }
                if (ok) {
                    init((fc = c).openOutputStream(), 1024, 0, 32 * 1024, 8 * 1024, 32 * 1024, true);
                } else {
                    c.close();
                }
            }
        } catch (Throwable e) {
            if (c != null) {
                try {
                    c.close();
                } catch (Throwable ee) {
                }
            }
        }
        if(!ok)throw new IOException();
    }

    private void def_reset() throws IOException {
        this.BTYPE = 2;
        this.status = ZipOutputStream.STREAM_INIT;
        for (int i = 0; i < HASHMAP_COUNT; i++) {
            this.HM[i] = new IntShortHashMap(2 * 1024);
        }
        lastBlock = false;

    }

    private void init(OutputStream outputStream, int size, int compressionType, int plainWindowSize, int huffmanWindowSize, int bufsize, boolean ZipFile) throws IOException {
        BufDataOutputStream(outputStream, bufsize);
        curMethod = 8;
        if (this.zipFile = ZipFile) {
            entries = new Vector();
        }
        curEntry = null;
        this.smallCodeBuffer = new int[2];
        this.inputBuffer = new byte[size + 300];
        this.litCount = new int[286];
        this.distCount = new int[30];
        this.plainDataWindow = new byte[(plainWindowSize / HASHMAP_COUNT) * HASHMAP_COUNT];
        this.lz77active = true;
        if (huffmanWindowSize < 1024 && huffmanWindowSize > 0) {
            huffmanWindowSize = 1024;
        }
        this.outputWindow = new byte[huffmanWindowSize];

        def_reset();
        if (GZIP) {
            // write GZIP header, if wanted 
            if (compressionType == ZipOutputStream.TYPE_GZIP) {
                /*
                +---+---+---+---+---+---+---+---+---+---+
                |ID1|ID2|CM |FLG|     MTIME     |XFL|OS | 
                +---+---+---+---+---+---+---+---+---+---+*/
                writebuf(31);
                writebuf(139);
                writebuf(8);
                writebuf(new byte[6]);
                writebuf(255);
            }
        }
    }

    /**
     * 
     * @param outputStream 		stream to write the compressed data in 
     * @param size	prefered 	size of the internal buffer
     * @param compressionType	TYPE_GZIP or TYPE_DEFLATE
     * @param plainWindowSize	this size is important for the lz77 search. Larger values
     * 		will result in better compression. Maximum is 32768.
     * @param huffmanWindowSize	this size is important for the huffmanencoding. A large
     * 		value will result to a better frequency statistic and therefore to a better compression.
     * @throws IOException 		might be thrown in case that the inputStream can not be read, 
     * 		the outputStream can not be written into or in case of wrong arguments
     */
    public ZipOutputStream(OutputStream outputStream, int size, int compressionType, int plainWindowSize, int huffmanWindowSize, int bufsize, boolean ZipFile) throws IOException {
        init(outputStream, size, compressionType, plainWindowSize, huffmanWindowSize, bufsize, ZipFile);
    }

    public void Def_close() throws IOException {

        this.Def_flush();

        // append the final tree, in case of dynamic trees
        if (this.BTYPE == 2) {

            // empty the huffmanwindow and force a new block on purpose, if there might
            // 		occur a new block before the last one. This is to make sure that the 
            //		final flag is set only for the "very" last one
            if ((this.outProcessed + 8 + (this.inEnd - this.inStart) * 8 / 3 > this.outputWindow.length)) {
                compileOutput();
            }

            // compile the few remaining bits into the final block
            LZ77(true);
            this.lastBlock = true;
            compileOutput();
        } else {
            // no final tree, just flush since there is just one single 
            //		and therefore final block
            LZ77(true);
        }

        writeFooter();


    }

    /**
     * It is strongly recomended NOT to call flush before close() since 
     *  close() is able to handle the flushing better itself.
     */
    public void Def_flush() throws IOException {
        // flush inputBuffer -> LZ77
        LZ77(false); // do not set to true, since we still need sth. for the last block

    }
    /* (non-Javadoc)
     * @see java.io.OutputStream#write(int)
     */

//    public void write(int b) throws IOException {
//
//        if (this.inputBuffer.length == this.inEnd) {
//            // process the inputBuffer if we need space 
//            LZ77(false);
//        }
//
//        // append the byte to the input buffer
//        this.inputBuffer[this.inEnd++] = (byte) b;
//
//        // ... and refresh the isize as well as crc32
//        this.isize++;
//        byte[] bb = new byte[1];
//        bb[0] = (byte) b;
//        this.crc32 = updateCRC(this.crc32, bb, 0, 1);
//
//    }
    public void Def_write(byte b[]) throws IOException {
        Def_write(b, 0, b.length);
    }

    public void Def_write(byte b[], int off, int len) throws IOException {
        int processed = 0;

        // refresh checksum
        this.crc = updateCRC(this.crc, b, off, len);
//        this.isize += len;

        while (processed != len) {

            // fill data in 
            if (this.inputBuffer.length - this.inEnd >= len - processed) {
                // if more than necessary fits in
                System.arraycopy(b, processed + off, this.inputBuffer, this.inEnd, len - processed);
                this.inEnd += len - processed;
                processed = len;
            } else {
                System.arraycopy(b, processed + off, this.inputBuffer, this.inEnd, this.inputBuffer.length - this.inEnd);
                processed += this.inputBuffer.length - this.inEnd;
                this.inEnd = this.inputBuffer.length;
            }

            // LZ77 the inputBuffer
            LZ77(false);
        }
    }

    /**
     * This function searches through all hashmaps and returns the best pointer
     * 		that was found.
     * 
     * @param bestPointer	this will hold the resulting distance and lenth pair {distance, length}
     * @param position		position to look at in the inputbuffer
     * @return				if there was a pointer found
     */
    private boolean search4LZ77(int[] bestPointer, int position) {
        IntShortHashMap.Element found = null;

        int[] pointer = new int[2];
        bestPointer[1] = 0;

        for (int i = 0; i < HASHMAP_COUNT; i++) {

            // retrieve and compare the best pointers out of each hashmap
            found = null;
            found = this.HM[i].get((128 + this.inputBuffer[position]) << 16 | (128 + this.inputBuffer[position + 1]) << 8 | (128 + this.inputBuffer[position + 2]));

            if (found != null && found.size != 0) {
                searchHM4LZ77(found, pointer, position);

                if (pointer[1] > bestPointer[1]) {
                    bestPointer[0] = pointer[0];
                    bestPointer[1] = pointer[1];
                }

            }
        }

        // was a pointer found?
        return bestPointer[1] != 0;
    }

    /**
     *  This function takes the result element of a hashmap to look for previous reoccurrences of the data in the inputbuffer
     *  starting at position. The best matching distance and length pair is return in
     *  pointer. 
     *  
     * @param found the hashmap result of the thee chars
     * @param pointer  distance and lenth pair {distance, length}
     * @param position  position in the inputbuffer
     */
    private void searchHM4LZ77(IntShortHashMap.Element found, int[] pointer, int position) {
        int length;

        // check this.plainPointer-found0.values[k] from the end since shortest pointers are at the end
        int bestK = 0, bestLength = 0;

        for (int k = found.size - 1; k >= 0; k--) {

            length = 3;

            int comparePointer = 100000;

            // deal with index out of bounds in case of finish
            while (length < 258 && position + length < this.inputBuffer.length) {
                if (found.values[k] < this.plainPointer) {
                    comparePointer = (found.values[k] + (length % (this.plainPointer - found.values[k]))) % this.plainDataWindow.length;
                } else {
                    comparePointer = (found.values[k] + (length % (this.plainPointer + this.plainDataWindow.length - found.values[k]))) % this.plainDataWindow.length;
                }

                // break, if the compared chars differ
                if (this.inputBuffer[position + length] == this.plainDataWindow[ comparePointer]) {
                    length++;
                } else {
                    break;
                }

            }

            // compare the recently found pointer pair with the currently best
            if (length > bestLength) {
                bestK = k;
                bestLength = length;
                if (length == 258) {
                    break;
                }
            }
        }

        // encode the pointer
        pointer[0] = /*distance=*/ (this.plainPointer - found.values[bestK] + this.plainDataWindow.length) % this.plainDataWindow.length;
        pointer[1] =/*length=*/ bestLength;

    }

    /**
     * This function puts the pointer into the outputwindow where it will be taken off
     * when the huffman compression takes place. 
     * 
     * @param distance		the distance information of the pointer
     * @param length		the length information of the pointer
     * @throws IOException 	this might happen, when there occurr errors while writing into the outputStream
     */
    private void encodePointer(int distance, int length) throws IOException {
        int litlen;
        byte litextra;
        int di;
        int distExtra;

        if (debug) {
            System.out.println(length + "-Tupel found  at " + distance + "back");
        }

        // compute length information
        di = encodeCode(LENGTH_CODE, length);
        litlen = 257 + di;
        litextra = (byte) (length - LENGTH_CODE[di * 2 + 1]);

        //compute distance information
        di = encodeCode(DISTANCE_CODE, distance);
        distExtra = distance - DISTANCE_CODE[di * 2 + 1];

        // write buffer information for compiler
        if (this.outputWindow.length != 0) {

            this.outputWindow[this.outProcessed] = (byte) 255; // special distance stuff
            this.outputWindow[this.outProcessed + 1] = (byte) (litlen - 255); // 0 is reserved!!

            this.outputWindow[this.outProcessed + 2] = litextra;// value of extra bits
            this.outputWindow[this.outProcessed + 3] = (byte) (di);// id of Table == code
            this.outputWindow[this.outProcessed + 4] = (byte) (distExtra & 0xFF);
            this.outputWindow[this.outProcessed + 5] = (byte) (distExtra >> 8 & 0xFF);
            this.outputWindow[this.outProcessed + 6] = (byte) (distExtra >> 16 & 0xFF);

            this.outProcessed += 6; // 7-1

            // increase distCount
            this.litCount[litlen]++;
            this.distCount[di]++;

        } else { // write direct when dynamic huffman is switched off

            // write litlen + extra bytes
            pushSmallBuffer(this.huffmanCode[litlen], this.huffmanCodeLength[litlen]);
            pushSmallBuffer(litextra, (byte) LENGTH_CODE[2 * (litlen - 257)]);

            // write distance code + extra info
            pushSmallBuffer(this.distHuffCode[di], this.distHuffCodeLength[di]);
            pushSmallBuffer(distExtra, (byte) DISTANCE_CODE[di * 2]);

        }

    }

    /**
     * This method fills a single char into the huffmanwindow or in the outputStream
     * 		depending on the huffmanmode
     *  
     * @param position		take the char from inputbuffer at this postion
     * @throws IOException	
     */
    private void encodeChar(int position) throws IOException {
        int val = (this.inputBuffer[position] + 256) & 255;

        if (this.outputWindow.length != 0) {
            // and increase the literal Count
            this.litCount[val]++;

            // fill the plain char in the output window ... encode 255
            this.outputWindow[this.outProcessed] = (byte) val;
            if (val == 255) {
                //special encoding for 255
                this.outProcessed++;
                this.outputWindow[this.outProcessed] = (byte) 0;
            }
        } else {
            pushSmallBuffer(this.huffmanCode[val], this.huffmanCodeLength[val]);
        }
    }

    /**
     * This function applies the LZ77 algorithm on the inputBuffer.
     * The inputBuffer is cleared completely and filled into 
     * outputWindow. If outputWindow is full, compileOutput is called.
     * @throws IOException 
     */
    //public int l=0;
    private void LZ77(boolean finish) throws IOException {

        // prepate the inputbuffer
        if (this.inStart != 0) {
            System.arraycopy(this.inputBuffer, this.inStart, this.inputBuffer, 0, this.inEnd - this.inStart);
            this.inEnd -= this.inStart;
            this.inStart = 0;
        }
        // do not leave sth. behind if finish is enabled
        int upTo;
        if (finish) {
            upTo = this.inEnd;
        } else {
            upTo = this.inEnd - 300;
        }

        // process the inputbuffer
        int pointer[] = new int[2];
        int lastpointer[] = new int[2];

        int distance;
        int length;

        int i;
        for (i = 0; i < upTo;) {

            length = 1;
            distance = 0;

            // TODO what about lazy matching??

            // search for reoccurrence, if there is enough input data left
            if (this.lz77active && i < upTo - 2 && search4LZ77(pointer, i)) {

                if (pointer[1] > lastpointer[1]) {
                    // put single char of (i-1) and keep searching
                    lastpointer[0] = pointer[0];
                    lastpointer[1] = pointer[1];
                } else {
                    distance = pointer[0];
                    length = pointer[1];
                }

            }

            // the pointer is not allowed to exceed the inputBuffer!!
            if (finish && upTo - i < length) {
                length = upTo - i;
            }

            // encode pointer or char
            if (length > 2) {
                encodePointer(distance, length);
            } else {
                encodeChar(i);
            }


            // check if Huffman compression is requested
            if (this.outputWindow.length != 0) {
                // check, if we have to compile the output
                this.outProcessed++;
                // there has to be enough space for a length, distance pair
                if (this.outProcessed + 8 > this.outputWindow.length) {
                    // if outputWindow full : call compileOutput
                    compileOutput();
                }

            }

            // refresh the plain window, if lz77 is active
            if (this.lz77active) {
                for (int k = 0; k < length; k++) {
                    this.plainDataWindow[this.plainPointer] = this.inputBuffer[i + k];

                    // add the bytes to the hashmap
                    this.HM[this.plainPointer / (this.plainDataWindow.length / HASHMAP_COUNT)].put((128 + this.inputBuffer[i + k]) << 16 | (128 + this.inputBuffer[i + k + 1]) << 8 | (128 + this.inputBuffer[i + k + 2]), (short) this.plainPointer);

                    // clear hashmap
                    if (++this.plainPointer % (this.plainDataWindow.length / HASHMAP_COUNT) == 0) {
                        // wrap around
                        if (this.plainPointer == this.plainDataWindow.length) {
                            this.plainPointer = 0;
                        }

                        this.HM[(this.plainPointer / (this.plainDataWindow.length / HASHMAP_COUNT)) % HASHMAP_COUNT].clear();
                    }

                }
            }
            // i is just increased here
            i += length;

        }

        this.inStart = i;
    }

    /**
     * This function generates and writes the header for a new block.
     * Therefore it has to calculate and compress the huffman trees.
     * 
     * @throws IOException
     */
    private void newBlock() throws IOException {
        if (this.status == ZipOutputStream.STREAM_INIT) {
            this.status = ZipOutputStream.STREAMING;
        } else {
            pushSmallBuffer(this.huffmanCode[256], this.huffmanCodeLength[256]);
            // write old 256, write blockheader
        }
        if (this.lastBlock) {
            pushSmallBuffer(1, (byte) 1);
            if (debug) {
                System.out.println("final block");
            }
        } else {
            pushSmallBuffer(0, (byte) 1);
        }

        pushSmallBuffer(this.BTYPE, (byte) 2);

        this.huffmanCode = new int[286];
        this.huffmanCodeLength = new byte[286];
        this.distHuffCode = new int[30];
        this.distHuffCodeLength = new byte[30];

        if (this.BTYPE == 1) {
            genFixedTree(this.huffmanCode, this.huffmanCodeLength, this.distHuffCode, this.distHuffCodeLength);
        } else if (this.BTYPE == 2) {


            // fake two distance codes, if there are none
            for (int i = 0; i < 2; i++) {
                if (this.distCount[i] == 0) {
                    this.distCount[i] = 1;
                }
            }

            // generate dynamic Huffmantree for Literals + Length
            this.litCount[256] = 1; // theoretisch kann man es auch auf 2 setzen, was das Problem jedoch nicht loest
            genTreeLength(this.litCount, this.huffmanCodeLength, 15);
            genHuffTree(this.huffmanCode, this.huffmanCodeLength);
            revHuffTree(this.huffmanCode, this.huffmanCodeLength);

            // generate dynamic Huffmantree for Distances
            genTreeLength(this.distCount, this.distHuffCodeLength, 15);
            genHuffTree(this.distHuffCode, this.distHuffCodeLength);
            revHuffTree(this.distHuffCode, this.distHuffCodeLength);

            // save tree
            compressTree(this.huffmanCodeLength, this.distHuffCodeLength);

            // clear the counter
            for (int i = 0; i < 286; i++) {
                this.litCount[i] = 0;
            }
            for (int i = 0; i < 30; i++) {
                this.distCount[i] = 0;
            }

        }

    }

    /**
     * This function applies the huffmanencoding on the collected data in outputWindow.
     * The huffman encoded data is then written into the outputStream.
     * 
     * @throws IOException 
     */
    private void compileOutput() throws IOException {
        if (debug) {
            System.out.println("  compile Output; new Block");
        }

        // generate & store the tree
        newBlock();

        int litlen, litextra, di, distExtra;

        // write the data
        int val = 0;
        for (int i = 0; i < this.outProcessed; i++) {

            val = this.outputWindow[i];
            if (val < 0) {
                val += 256;
            }
            if (val != 255) {
                pushSmallBuffer(this.huffmanCode[val], this.huffmanCodeLength[val]);
            } else {
                if (val == 255) {
                    i++;
                    if (this.outputWindow[i] == 0) {
                        // 255 char
                        pushSmallBuffer(this.huffmanCode[255], this.huffmanCodeLength[255]);

                    } else if (this.outputWindow[i] > 0) {
                        // compile encoded pointer

                        // length information
                        litlen = 255 + this.outputWindow[i];
                        i++;
                        litextra = this.outputWindow[i];
                        i++;

                        // distance information
                        di = this.outputWindow[i];
                        i++;
                        distExtra = ((this.outputWindow[i] + 256) & 255) | ((this.outputWindow[i + 1] + 256) & 255) << 8 | ((this.outputWindow[i + 2] + 256) & 255) << 16;
                        i += 3;
                        //distance=distExtra + ZipUtil.DISTANCE_CODE[di*2+1];

                        // write litlen + extra bytes
                        pushSmallBuffer(this.huffmanCode[litlen], this.huffmanCodeLength[litlen]);
                        pushSmallBuffer(litextra, (byte) LENGTH_CODE[2 * (litlen - 257)]);

                        // write distance code + extra info
                        pushSmallBuffer(this.distHuffCode[di], this.distHuffCodeLength[di]);
                        pushSmallBuffer(distExtra, (byte) DISTANCE_CODE[di * 2]);

                        i--;
                    } else {
                        throw new IOException("illegal code decoded");
                    }

                }

            }
        }
        this.outProcessed = 0;
    }

    /**
     * This function finishes the last block and writes the checksums.
     * 
     * @throws IOException
     */
    private void writeFooter() throws IOException {
        // write current 256
        pushSmallBuffer(this.huffmanCode[256], this.huffmanCodeLength[256]);
        if (debug) {
            System.out.println(" wrote final 256;");
        }

        // flush the upto the byte boundrary
        if ((this.smallCodeBuffer[1] & 7) != 0) {
            pushSmallBuffer(0, (byte) (8 - (this.smallCodeBuffer[1] & 7)));
        }


        if (!zipFile) {
            //write CRC, count
            this.writeLint(crc);
            this.writeLint(size);
            // close the stream
            flushbuf();
            close();
        }
        if (debug) {
            System.out.println(" output finished");
        }
    }

    /**
     * This function compresses and writes the literal/length as well as the distance trees. 
     * At first repeatcodes are determined, then the miniTree is generated and after all they
     * are written into the outputStream.
     * 
     * @param huffmanCodeLength
     * @param distHuffCodeLength
     * @throws IOException
     */
    private void compressTree(byte[] huffmanCodeLength, byte[] distHuffCodeLength) throws IOException {

        int HLIT = 285;
        int HDIST = 29;

        while (huffmanCodeLength[HLIT] == 0 && HLIT > 29) {
            HLIT--;
        }
        HLIT++;

        // dont worry: empty distance tree workaround is in newBlock();
        while (distHuffCodeLength[HDIST] == 0 && HDIST > 0) {
            HDIST--;
        }
        HDIST++; // # of all Distance Symbols

        // merge Hlit + Hdist
        byte[] len = new byte[HLIT + HDIST];
        int j = 0;
        for (int i = 0; i < HLIT; i++) {
            len[j] = huffmanCodeLength[i];
            //System.out.println("  " + j + " " + huffmanCode[j] + ", " + huffmanCodeLength[j]);
            j++;
        }
        for (int i = 0; i < HDIST; i++) {
            len[j] = distHuffCodeLength[i];
            //System.out.println("  " + j + " " + distHuffCode[i] + ", " + distHuffCodeLength[i]);
            j++;
        }

        int[] miniHuffData = {16, 17, 18, 0, 8, 7, 9, 6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15};

        // fill repeatcodes in and count them
        byte[] outLitLenDist = new byte[HLIT + HDIST];
        int outCount = 0;
        int[] miniCodeCount = new int[19]; // count of used miniCodes
        short k;
        for (int i = 0; i < len.length; i++) {

            if (i + 3 < len.length && len[i] == len[i + 1] && len[i] == len[i + 2] && len[i] == len[i + 3]) {
                if (len[i] == 0) {
                    outLitLenDist[outCount] = 0; // ausgliedern
                    //System.out.println("wrinting outCount="+outCount + "   =  0");
                    k = 4;

                    while (i + k < len.length && len[i] == len[i + k] && k < 139) {
                        k++;
                    }


                    if (k < 11 + 1) {
                        //System.out.println("writing outCount="+outCount + "   =  17");
                        outLitLenDist[outCount + 1] = 17;
                        outLitLenDist[outCount + 2] = (byte) (k - 3 - 1);
                    } else {
                        //System.out.println("writing outCount="+outCount + "   =  18");
                        //System.out.println(" repeat count = " + (k-1));
                        outLitLenDist[outCount + 1] = 18;
                        outLitLenDist[outCount + 2] = (byte) (k - 11 - 1);
                    }

                    i += k - 1;
                } else {
                    // encode repeat code ; char * n -> char, 16, n
                    //System.out.println("writing outCount="+outCount + "   =  "+len[i]);
                    outLitLenDist[outCount] = len[i];
                    outLitLenDist[outCount + 1] = 16;
                    //System.out.println("writing outCount="+outCount + "   =  16");

                    // repeat count
                    k = 4;
                    while (i + k < len.length && len[i] == len[i + k] && k < 7) {
                        k++;
                    }
                    outLitLenDist[outCount + 2] = (byte) (k - 4);
                    i += k - 1;
                }
                // add value to statistic
                miniCodeCount[outLitLenDist[outCount]]++;
                // add repeat code to statistic
                miniCodeCount[outLitLenDist[outCount + 1]]++;
                // +3  == 2 +1
                outCount += 2;

            } else {
                //System.out.println("wrinting outCount="+outCount + "   =  "+ len[i]);
                outLitLenDist[outCount] = len[i];
                miniCodeCount[outLitLenDist[outCount]]++;
            }

            outCount++;
        }


        // generate mini Tree
        byte[] miniHuffCodeLength = new byte[19];
        int[] miniHuffCode = new int[19];

        int i = 0;

        genTreeLength(miniCodeCount, miniHuffCodeLength, 7);

        genHuffTree(miniHuffCode, miniHuffCodeLength);
        revHuffTree(miniHuffCode, miniHuffCodeLength);

        // write Header-Header
        pushSmallBuffer(HLIT - 257, (byte) 5);
        pushSmallBuffer(HDIST - 1, (byte) 5);


        int HCLEN = 19 - 1;
        while (miniHuffCodeLength[miniHuffData[HCLEN]] == 0 && HCLEN > 0) {
            HCLEN--;
        }
        HCLEN++;

        pushSmallBuffer(HCLEN - 4, (byte) 4);

        // write Mini-Tree
        for (i = 0; i < HCLEN; i++) {
            //System.out.println("" + i + "=" +miniHuffData[i] + ":\t" + miniHuffCodeLength[miniHuffData[i]]);
            pushSmallBuffer(miniHuffCodeLength[miniHuffData[i]], (byte) 3);

        }
        if (debug) {
            System.out.println(" HLIT: " + HLIT);
            System.out.println(" HDIST: " + HDIST);
            System.out.println(" HCLEN: " + HCLEN);
        }

        // write Huffmanntree
        for (i = 0; i < outCount; i++) {
            // write code
            //System.out.println("wrinting i="+i + "   =  " + outLitLenDist[i]);
            pushSmallBuffer(miniHuffCode[outLitLenDist[i]], miniHuffCodeLength[outLitLenDist[i]]);

            // add special treatment
            if (outLitLenDist[i] > 15) {
                switch (outLitLenDist[i]) {
                    case 16:
                        pushSmallBuffer(outLitLenDist[i + 1], (byte) 2);
                        i++;
                        break;
                    case 17:
                        pushSmallBuffer(outLitLenDist[i + 1], (byte) 3);
                        i++;
                        break;
                    default: // 18
                        pushSmallBuffer(outLitLenDist[i + 1], (byte) 7);
                        i++;
                        break;
                }

            }

            //System.out.println(outLitLenDist[i] + " == code: " + miniHuffCode[outLitLenDist[i]]);
        }

    }

    /**
     * This function is able to write bits into the outStream. Please
     * mind that is uses a buffer. You might have to flush it, by giving zeros.
     *   
     * @param val The bitsequence as an integer.
     * @param len The number of bits to process.
     * @throws IOException in case Errors within the outputStream ocurred
     */
    private void pushSmallBuffer(int val, byte len) throws IOException {
        int smallBuffer0 = this.smallCodeBuffer[0];
        int smallBuffer1 = this.smallCodeBuffer[1];

        // add the given data
        smallBuffer0 &= ~(((1 << len) - 1) << smallBuffer1);
        smallBuffer0 |= (val << smallBuffer1);
        smallBuffer1 += len;

        // clear the buffer except for a fraction of a reamining byte
        while (smallBuffer1 >= 8) {
            writebuf(smallBuffer0 & 255);
            smallBuffer0 >>>= 8;
            smallBuffer1 -= 8;
        }

        this.smallCodeBuffer[0] = smallBuffer0;
        this.smallCodeBuffer[1] = smallBuffer1;
    }
    /////////////////////////helper
    // constants:
     static final int[] LENGTH_CODE = {
        0, 3,
        0, 4,
        0, 5,
        0, 6,
        0, 7,
        0, 8,
        0, 9,
        0, 10,
        1, 11,
        1, 13,
        1, 15,
        1, 17,
        2, 19,
        2, 23,
        2, 27,
        2, 31,
        3, 35,
        3, 43,
        3, 51,
        3, 59,
        4, 67,
        4, 83,
        4, 99,
        4, 115,
        5, 131,
        5, 163,
        5, 195,
        5, 227,
        0, 258};
     static final int[] DISTANCE_CODE = {
        0, 1,
        0, 2,
        0, 3,
        0, 4,
        1, 5,
        1, 7,
        2, 9,
        2, 13,
        3, 17,
        3, 25,
        4, 33,
        4, 49,
        5, 65,
        5, 97,
        6, 129,
        6, 193,
        7, 257,
        7, 385,
        8, 513,
        8, 769,
        9, 1025,
        9, 1537,
        10, 2049,
        10, 3073,
        11, 4097,
        11, 6145,
        12, 8193,
        12, 12289,
        13, 16385,
        13, 24577};

     static int encodeCode(int[] Code, int distance) {
        int i = 0;
        //while(i<Code.length/2 && distance>=Code[i*2+1]){i++;}
        while (i < (Code.length >> 1) && distance >= Code[(i << 1) + 1]) {
            i++;
        }
        // now: distance < Code[i*2+1]

        return i - 1;
    }

    /**
     * This function generates the huffmann codes according to a given set of bitlegth.
     * 
     * @param huffmanCode			will be the resulting code pattern
     * @param huffmanCodeLength		has to be the bitlegth for each code
     */
    static void genHuffTree(int[] huffmanCode, byte[] huffmanCodeLength) {
        int maxbits = 0;

        // generate bitlen_count
        for (int i = 0; i < huffmanCodeLength.length; i++) {
            int length = huffmanCodeLength[i];
            maxbits = maxbits > length ? maxbits : length;
        }
        maxbits++;

        short[] bitlen_count = new short[maxbits];
        for (int i = 0; i < huffmanCodeLength.length; i++) {
            bitlen_count[huffmanCodeLength[i]]++;
        }

        int code = 0;
        int[] next_code = new int[maxbits];
        bitlen_count[0] = 0;

        // find the Codes for each smallest Code within all of the same bitlength
        for (int bits = 1; bits < maxbits; bits++) {
            //code=(code + bitlen_count[bits-1])*2;
            code = (code + bitlen_count[bits - 1]) << 1;
            next_code[bits] = code;
        }
        // generate all codes by adding 1 to the predecessor
        for (int i = 0; i < huffmanCode.length; i++) {
            byte length = huffmanCodeLength[i];
            if (length != 0) {
                huffmanCode[i] = next_code[length];
                next_code[length]++;
            }
        }

    }

    static void revHuffTree(int[] huffmanCode, byte[] huffmanCodeLength) {
        // reverse all:
        int tmp;
        int reversed;
        for (int i = 0; i < huffmanCode.length; i++) {
            tmp = huffmanCode[i];
            reversed = 0;
            for (int j = 0; j < huffmanCodeLength[i]; j++) {
                reversed |= ((tmp >>> j) & 1);
                reversed <<= 1;
            }
            huffmanCode[i] = reversed >>> 1;
        }
    }

    /**
     * Generate the fixed huffman trees as described in the rfc 
     * 
     * @param huffmanCode - pointer to where the code shall be inserted
     * @param huffmanCodeLength
     * @param distHuffCode
     * @param distHuffCodeLength
     */
     static void genFixedTree(int[] huffmanCode, byte[] huffmanCodeLength,
            int[] distHuffCode, byte[] distHuffCodeLength) {

        int i;
        // huffmanCodes
        for (i = 0; i <= 143; i++) {
            huffmanCode[i] = 48 + i; //48==00110000
            huffmanCodeLength[i] = 8;
        }
        for (i = 144; i <= 255; i++) {
            huffmanCode[i] = 400 + i - 144; //400==110010000
            huffmanCodeLength[i] = 9;
        }
        for (i = 256; i <= 279; i++) {
            huffmanCode[i] = i - 256; //0==0
            huffmanCodeLength[i] = 7;
        }
        for (i = 280; i < 286; i++) {
            huffmanCode[i] = 192 + i - 280; //192==11000000
            huffmanCodeLength[i] = 8;
        }
        // reverse all:
        revHuffTree(huffmanCode, huffmanCodeLength);

        // distHuffCode for non fixed DISTANCE tree
        for (int j = 0; j < distHuffCode.length; j++) {
            distHuffCode[j] = j;
            distHuffCodeLength[j] = 5;
        }
        // reverse all:
        revHuffTree(distHuffCode, distHuffCodeLength);
    }

    /**
     * Compute the tree length according to the frequency of each code 
     * 
     * @param count 
     * @param huffmanCodeLength
     * @param max_len 
     */
     static void genTreeLength(int[] count, byte[] huffmanCodeLength, int max_len) {
        int[] knotCount = new int[count.length];
        int[] knotPointer = new int[count.length];

        // initialise the knots
        for (short i = 0; i < count.length; i++) {
            if (count[i] != 0) {
                knotCount[i] = count[i];
            } else {
                knotCount[i] = Integer.MAX_VALUE;
            }
            knotPointer[i] = i;
        }

        // look for the smalest two knots
        int s1 = 0, s2 = 0;
        while (true) {

            // find Min1 and Min2
            if (knotCount[0] < knotCount[1]) {
                s1 = 0;
                s2 = 1;
            } else {
                s1 = 1;
                s2 = 0;
            }
            for (int i = 2; i < count.length; i++) {
                if (knotCount[i] < knotCount[s1]) {
                    s2 = s1;
                    s1 = i;
                } else if (knotCount[i] < knotCount[s2]) {
                    s2 = i;
                }
            }

            // if we got all Minima -> exit
            if (knotCount[s2] == Integer.MAX_VALUE) {
                break;
            }

            // merge the knots
            knotCount[s1] += knotCount[s2];
            int tmp = knotPointer[s2];
            knotCount[s2] = Integer.MAX_VALUE;

            // set all knot pointer of Min2 -> Min1,   add one to CodeLength, and remove old knots from search
            for (int i = 0; i < count.length; i++) {
                if (knotPointer[i] == tmp) {
                    knotPointer[i] = knotPointer[s1];
                    huffmanCodeLength[i]++;
                } else if (knotPointer[i] == knotPointer[s1]) {
                    huffmanCodeLength[i]++;
                }
            }

        }


        // check for overflow
        int overflowCount = 0;
        for (int i = 0; i < huffmanCodeLength.length; i++) {
            if (huffmanCodeLength[i] > max_len) {
                overflowCount++;
            }
        }
        // fix the overflow
        if (overflowCount != 0) {
            //System.out.println(" fixing " + overflowCount + " overflows  because of max=" + max_len);

            // list the overflows
            short overflows[] = new short[overflowCount];
            overflowCount = 0;
            for (short i = 0; i < huffmanCodeLength.length; i++) {
                if (huffmanCodeLength[i] > max_len) {
                    overflows[overflowCount++] = i;
                }
            }
            // find the index of the smalest node
            int minNode = 0;
            for (int i = 0; i < huffmanCodeLength.length; i++) {
                if (huffmanCodeLength[i] != 0 && huffmanCodeLength[minNode] > huffmanCodeLength[i]) {
                    minNode = i;
                }
            }

            // ok lets go and fix it....
            int exendableNode;
            int overflow1, overflow2;
            while (overflowCount != 0) {
                exendableNode = minNode;
                // find the largest expandable length
                for (int i = 0; i < huffmanCodeLength.length; i++) {
                    if (huffmanCodeLength[i] < max_len && huffmanCodeLength[exendableNode] < huffmanCodeLength[i]) {
                        exendableNode = i;
                    }
                }

                // find the deepest two overflows
                overflow1 = 0;
                overflow2 = 0;
                for (int i = 0; i < overflows.length; i++) {
                    if (huffmanCodeLength[overflows[i]] > huffmanCodeLength[overflow1]) {
                        overflow1 = overflows[i];

                    } else if (huffmanCodeLength[overflows[i]] == huffmanCodeLength[overflow1]) {
                        overflow2 = overflows[i];
                    }
                }

                // insert one of them at the best exendableNode
                huffmanCodeLength[exendableNode]++;
                huffmanCodeLength[overflow1] = huffmanCodeLength[exendableNode];

                // lift the other one
                huffmanCodeLength[overflow2]--;

                // refresh the overflow count
                overflowCount--;
                if (huffmanCodeLength[overflow2] == max_len) {
                    overflowCount--;
                }
            }
        }

    }

    /**
     *  This function converts the representation of a huffman tree from
     *  table to tree structure.
     *  
     * @param huffmanCode	input
     * @param huffmanCodeLength		input
     * @param huffmanData	input
     * @param huffmanTree   resulting tree
     */
     static void convertTable2Tree(int[] huffmanCode, byte[] huffmanCodeLength, int[] huffmanData, short[] huffmanTree) {

        for (int i = 0; i < huffmanTree.length; i++) {
            huffmanTree[i] = 0;
        }

        short pointer;
        short nextNode = 1;
        // fill each code in
        short j;
        for (short i = 0; i < huffmanCode.length; i++) {
            if (huffmanCodeLength[i] != 0) {
                pointer = 0;
                for (j = 0; j < huffmanCodeLength[i]; j++) {

                    if (huffmanTree[pointer * 2] == 0) {
                        // undefined internal NODE
                        huffmanTree[pointer * 2] = nextNode++;
                        huffmanTree[pointer * 2 + 1] = nextNode++;
                    }
                    // go to known internal NODE
                    pointer = huffmanTree[pointer * 2 + ((huffmanCode[i] >>> j) & 1)];

                }

                // set empty LEAF to data
                if (pointer < 0) {
                    if (debug) {
                        System.out.println("error pointer=-1");
                    }
                }
                huffmanTree[pointer * 2] = -1;
                huffmanTree[pointer * 2 + 1] = (short) huffmanData[i];
            }
        }
    }

    /**
     * This function parses and pops one huffmancode out of the given buffer
     * and returs the associated value.
     * 
     * @param smallCodeBuffer the buffer that is parsed
     * @param huffmanTree is the representation of the tree
     * @return the data represented by the parsed huffman code
     * @throws IOException
     */
     static int deHuffNext(long[] smallCodeBuffer, short[] huffmanTree) throws IOException {
        if (debug) {
            System.out.println("deHuffNext(long[], short[])");
        }
        if (debug) {
            if (smallCodeBuffer[1] < 15) {
                System.out.println("smallCodebuffer is too small");
            }
        }

        short pointer = 0;

        while (huffmanTree[pointer * 2] != -1) {
            // go to next internal NODE
            pointer = huffmanTree[pointer * 2 + (int) (smallCodeBuffer[0] & 1)];
            smallCodeBuffer[0] >>>= 1;
            smallCodeBuffer[1]--;

            if (pointer == 0) {
                //System.out.println("small code buffer DAta "+smallCodeBuffer[0] + "\n");
                // error: in walking tree no leaf found
                throw new IOException("5");
            }

        }
        // return the data
        return huffmanTree[pointer * 2 + 1];

    }
    ////////////////////////////Zip
    private Vector entries;
    private int crc;
    private ZipEntry curEntry;
    private int curMethod, size, entryoffset;

    public void putNextEntry(String name) throws Exception {
        ZipEntry entry = new ZipEntry(name);
        entry.setTime(System.currentTimeMillis());
        putNextEntry(entry);
    }

    public void putNextEntry(ZipEntry entry) throws Exception {
        if (curEntry != null) {
            closeEntry();
        }
        entry.flag = 8;
        entry.setMethod(curMethod);
        curEntry = entry;
        entries.addElement(entry);
        entry.offset = getoffse();
        writeLint(0x04034b50);//4
        writeLshort(entry.getMethod() == 0 ? 10 : 20);//6
        writeEntry(entry);
        writebuf(entry.getNamebytes());
        crc = size = 0;
        def_reset();
        entryoffset = getoffse();
    }

    public void putFooEntry(ZipEntry entry,ZipFile zip) throws Exception {
        if (curEntry != null) {
            closeEntry();
        }
        zip.setOffset(entry.offset+26);
        zip.skipbuf(zip.readLshort()+zip.readLshort());
        entry.flag = 0;
        entries.addElement(entry);
        entry.offset = getoffse();
        writeLint(0x04034b50);//4
        /**
         * version made by (2 bytes)
         * the upper byte indicates the compatibility of the file
         * attribute information. if the external file attributes
         * are compatible with ms-dos and can be read by pkzip for
         * dos version 2.04g then this value will be zero. if these
         * attributes are not compatible, then this value will identify
         *  the host system on which the attributes are compatible.
         *  software can use this information to determine the line
         *  record format for text files etc. the current
         *  mappings are:
         *  0 - ms-dos and os/2 (fat / vfat / fat32 file systems)
         *  1 - amiga 2 - vax/vms
         *  3 - unix 4 - vm/cms
         *  5 - atari st 6 - os/2 h.p.f.s.
         *  7 - macintosh 8 - z-system
         *  9 - cp/m 10 - windows ntfs
         * 11 thru 255 - unused
         * the lower byte indicates the version number of the
         * software used to encode the file. the value/10
         * indicates the major version number, and the value
         * mod 10 is the minor version number.
        
         */
        writeLshort(entry.isDirectory() ? 10 : 20);//6
        writeEntry(entry);
        writebuf(entry.getNamebytes());
        writebuf(zip.bytes,zip.getOffst(),entry.compressedSize);
    }

    public void write(byte[] b) throws IOException {
        write(b, 0, b.length);
    }

    public void closeEntry() throws Exception {

        if (curEntry == null) {
            return;
        }
        if (curMethod == 8) {
            Def_close();
        }
        ZipEntry entry = curEntry;
        entry.size = size;
        int now = getoffse();
        entry.compressedSize = now - entryoffset;
        entry.crc = crc;
        if (!resetTo(entry.offset + 6)) {
            //   System.out.println("resetTo " + (entry.offset + 6) + "  fail");
            writeLint(0x08074b50);
            writeLint(entry.crc);
            writeLint(entry.compressedSize);
            writeLint(entry.size);
        } else {
            //   System.out.println("resetTo " + (entry.offset + 6) + "  ok");
            entry.flag = 0;
            writeEntry(entry);
            skipto(now);
        }
        flushbuf();
        curEntry = null;
    }

    public void write(byte b[], int off, int len) throws IOException {
        if (curMethod == 0) {
            writebuf(b, off, len);
            crc = updateCRC(crc, b, off, len);
        } else {
            Def_write(b, off, len);
        }
        size += len;
    }

    private void writeEntry(ZipEntry entry) throws IOException {
        writeLshort(entry.flag);//8           General purpose bit flag
        writeLshort(entry.getMethod());//10   Compression method
        writeLint(entry.dostime);//14         Last mod file time and data
        writeLint(entry.crc);//18             CRC-32
        writeLint(entry.compressedSize);//22  Compressed size
        writeLint(entry.size);//26            Uncompressed size
        writeLshort(entry.namelen);//28       Filename length (f)
        writeLshort(0);//30 extralen          Extra field length (e)
    }

    private void finish() throws Exception {

        if (curEntry != null) {
            closeEntry();
        }
        int dataoff = getoffse();
        int sizeEntries = 0;
        ZipEntry entry;
        int numEntries = entries.size();
        for (int k = 0; k < numEntries; ++k) {
            entry = (ZipEntry) entries.elementAt(k);
            //  int method = entry.getMethod();
            writeLint(0x02014b50);   //           Central file header signature
            writeLshort(20); //Version made by
            writeLshort(entry.isDirectory() ? 10 : 20);  //                  Version needed to extract
            writeEntry(entry);
            writeLshort(0);//                      File comment length (c)
            writeLshort(0);            // Disk number start
            /**
             * the lowest bit of this field indicates, if set, that
             * the file is apparently an ascii or text file. if not
             * set, that the file apparently contains binary data.
             * the remaining bits are unused in version 1.0.
             */
            //   writeLshort(entry.name.endsWith(".txt") || entry.name.endsWith(".log")||entry.name.endsWith(".java")||entry.name.endsWith(".cfg") ? 1 : 0);            // Internal file attributes
            writeLshort(0);
            /**
             * the mapping of the external attributes is
             * host-system dependent (see 'version made by'). for
             * ms-dos, the low order byte is the ms-dos directory
             * attribute byte. if input came from standard input, this
             * field is set to zero.
             */
            writeLint(entry.isDirectory() ? 1103101968 : -2122317792);                      //  External file attributes
            writeLint(entry.offset);           //  Relative offset of local header
            writebuf(entry.getNamebytes());
            sizeEntries += 46 + entry.namelen;
        }
        writeLint(0x06054b50);//4         End of central dir signature (0x06054b50)
        writeLint(0);         //          Number of this disk and Number of the disk with the start of the central directory

        writeLshort(numEntries);//        Total number of entries in the central dir on this disk
        writeLshort(numEntries);//12      Total number of entries in the central dir
        writeLint(sizeEntries);//16       Size of the central directory
        writeLint(dataoff); //             Offset of start of central directory with respect to the starting disk number
        writeLshort(0);   //              zipfile comment length (c)
        flushbuf();
        entries = null;
    }

    public void close() {
        try {
            if (zipFile && entries != null) {
                finish();
            }else{
            flushbuf();
            }
            bytes = null;
            outbuf.close();
            if (fc != null) {
                fc.close();
            }
        } catch (Throwable ex) {
        }
        this.outputWindow = null;
        this.inputBuffer = null;
        this.litCount = null;
    }
    ///////////////////////boutbuf
    OutputStream outbuf;
    byte[] bytes;
    private int bufoff;
    private int offset;

    private void BufDataOutputStream(OutputStream output, int bufsize) {
        outbuf = output;
        bytes = new byte[bufsize];
        bufoff = 0;
        offset = 0;
    }

    private boolean resetTo(int off) {
        int l = offset - off;
        if (l <= bufoff) {
            bufoff -= l;
            offset = off;
            return true;
        }
        return false;
    }

    private int getoffse() {
        return offset;
    }

    private void skipto(int off) {
        bufoff += off - offset;
        offset = off;
    }

    private void flushbuf() throws IOException {
        if (bufoff > 0) {
            outbuf.write(bytes, 0, bufoff);
            bufoff = 0;
        }
    }

    private void writebuf(byte[] addr, int off, int len) throws IOException {
        if (len <= 0) {
            return;
        }
        if (len < bytes.length - bufoff) {
            System.arraycopy(addr, off, bytes, bufoff, len);
            bufoff += len;
        } else {
            flushbuf();
            outbuf.write(addr, off, len);
        }
        offset += len;
    }

    private void writebuf(byte[] b) throws IOException {
        writebuf(b, 0, b.length);
    }

    private void writebuf(int l) throws IOException {
        if (bytes.length < bufoff + 1) {
            flushbuf();
        }
        ++offset;
        bytes[bufoff++] = (byte) l;
    }

    private void writeLshort(int l) throws IOException {
        if (bytes.length < bufoff + 2) {
            flushbuf();
        }
        offset += 2;
        bytes[bufoff++] = (byte) l;
        bytes[bufoff++] = (byte) (l >> 8);
    }

    private void writeLint(int l) throws IOException {
        if (bytes.length < bufoff + 4) {
            flushbuf();
        }
        offset += 4;
        bytes[bufoff++] = (byte) l;
        bytes[bufoff++] = (byte) (l >> 8);
        bytes[bufoff++] = (byte) (l >> 16);
        bytes[bufoff++] = (byte) (l >> 24);
    }
}
